exception Unimplemented

type exp = Int of int | Var of string 
       | Plus of exp * exp | Times of exp * exp
type stmt = Skip | Assign of string * exp
          | Seq of stmt * stmt 
          | If of exp * stmt * stmt
          | While of exp * stmt

let rec xlate_e (e:exp) =
  match e with
   Int i       ->  (fun h -> i)
  |Var str     ->  (fun h -> h str)
  |Plus(e1,e2) ->  let f1 = xlate_e e1 in
                   let f2 = xlate_e e2 in
                   (fun h -> (f1 h) + (f2 h))
  |Times(e1,e2) -> let f1 = xlate_e e1 in
                   let f2 = xlate_e e2 in
                   (fun h -> (f1 h) * (f2 h))

(* an example *)
let e = Plus(Int 3, Times(Var "x", Int 4))
let f = xlate_e e (* compile *)
(* the value bound to f is a function whose body does not use any IMP abstract syntax! *)
let ans = f (fun s -> 0)(* run w/ empty heap *)

let rec xlate_s (s:stmt) =
 match s with
  Skip          -> (fun h -> h)
 |Assign(str,e) -> 
   let f = xlate_e e in
   (fun h -> let i = f h in
             (fun s -> if s=str then i else h s))
 |Seq(s1,s2) ->
    let f2 = xlate_s s2 in (* order irrelevant! *)
    let f1 = xlate_s s1 in
    (fun h -> f2 (f1 h)) (* order relevant *)
|If(e,s1,s2) -> 
   let f1 = xlate_s s1 in
   let f2 = xlate_s s2 in
   let f  = xlate_e e  in
   (fun h -> if (f h) <> 0 then f1 h else f2 h)
|While(e,s1) ->
   let f1 = xlate_s s1 in
   let f  = xlate_e e  in
   let rec loop h = (* ah, recursion! *)
     if f h <> 0
     then loop (f1 h)
     else h
   in loop

let interp_prog s = 
  ((xlate_s s) (fun str -> 0)) "ans"

type stmt2 = Skip | Assign of string * exp
          | Seq of stmt2 * stmt2 
          | If of exp * stmt2 * stmt2
          | While of exp * stmt2
	  | Call1 of string * exp
	  | Call2 of string * exp

type funs = (string*(string*stmt2)) list
type prog = funs * stmt2

type heap = (string * int) list

let rec lookup h str =
  match h with
    [] -> 0 (* kind of a cheat *)
  |(s,i)::tl -> if s=str then i else lookup tl str

let update h str i = (str,i)::h

let rec interp_e (h:heap) (e:exp) =
 match e with
  Int i       ->i
 |Var str     ->lookup h str
 |Plus(e1,e2) ->(interp_e h e1)+(interp_e h e2)
 |Times(e1,e2)->(interp_e h e1)*(interp_e h e2)

let fresh h s = raise Unimplemented

let rec interp_s (fs:funs) (h:heap) (s:stmt2) =
  match s with
   Skip -> h
  |Seq(s1,s2) -> let h2 = interp_s fs h s1 in 
                 interp_s fs h2 s2
  |If(e,s1,s2) -> if (interp_e h e) <> 0
                  then interp_s fs h s1 
                  else interp_s fs h s2
  |Assign(str,e) -> update h str (interp_e h e)
  |While(e,s1) -> if (interp_e h e) <> 0
                  then let h2 = interp_s fs h s1 in
                       interp_s fs h2 s
                  else h
  | Call1(str,e) -> 
      let (arg,body) = List.assoc str fs in
    (* str(e) becomes arg:=e; body *)
      interp_s fs h (Seq(Assign(arg,e),body))
  | Call2(str,e) -> 
    let (arg,body) = List.assoc str fs in
    let y = fresh h s in
  (* str(e) becomes y:=arg; arg:=e; body; arg:=y
     where y is fresh *)
    interp_s fs h (Seq(Assign(y,Var arg),
                   Seq(Assign(arg,e),
                   Seq(body,
                       Assign(arg,Var y)))))
